import logging
import logging.config
import pathlib
import yaml
import json
from datetime import datetime
from typing import Union


iso8601_datefmt = "%Y-%m-%dT%H:%M:%S.%f"


# Logging matters
class UTC_ISO8601_Formatter(logging.Formatter):
    converter = datetime.utcfromtimestamp

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            s = ct.strftime(iso8601_datefmt)
        return s


log_message_format = (
    "%(asctime)s\t"
    "%(levelname)s\t"
    "%(name)s.%(funcName)s@%(lineno)d:\t"
    "%(message)s\t"
    "EOL"
)

log_formatter = UTC_ISO8601_Formatter(fmt=log_message_format)


def initialize_logger(
    level=logging.DEBUG,
    fmt: Union[logging.Formatter, str] = None,
    config: Union[pathlib.Path, str, dict] = None,
):
    # Initialize to good (tm) defaults
    logging.basicConfig(
        level=level,
        format=log_message_format,
        datefmt=iso8601_datefmt
    )

    if not fmt:
        fmt = log_message_format

    if not isinstance(fmt, logging.Formatter):
        msg_formatter = UTC_ISO8601_Formatter(fmt=str(fmt))

    # Override with explicit opthions
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    root_logger.handlers[0].setFormatter(msg_formatter)

    # Override with config options
    if config:
        if not isinstance(config, dict):
            conf_file = pathlib.Path(str(config))
            suffix = conf_file.suffix.lower()
            config = None

            if not conf_file.is_file():
                raise FileNotFoundError(conf_file)

            # assume yaml, unelss explicitly otherwise
            if not config and suffix not in ['.ini', '.json']:
                logging.getLogger(__name__).debug(
                    f"Attempting to parse '{conf_file}' as yaml file."
                )
                try:
                    with conf_file.open() as f:
                        config = yaml.safe_load(f.read())
                        logging.getLogger(__name__).info(
                            f"Successfully read '{conf_file}''"
                        )
                except Exception as ex:
                    if suffix in ['.yaml', '.yml']:
                        raise ex
                    else:
                        logging.getLogger(__name__).warning(
                            f"Unable to parse '{conf_file}' as yaml file."
                        )

            # if not yaml, assume json, unless explicitly otherwise
            if suffix not in ['.ini', '.yaml', '.yml']:
                logging.getLogger(__name__).debug(
                    f"Attempting to parse '{conf_file}' as json file."
                )
                try:
                    with conf_file.open() as f:
                        config = json.load(f)
                        logging.getLogger(__name__).info(
                            f"Successfully read '{conf_file}'"
                        )

                except Exception as ex:
                    if suffix == '.json':
                        raise ex
                    else:
                        logging.getLogger(__name__).warning(
                            f"Unable to parse '{conf_file}' as json file."
                        )

            # if neither json nor yaml, attempt to parse as legacy ini
            if not config and suffix not in ['.json', '.yaml', '.yml']:
                logging.getLogger(__name__).debug(
                    f"Attempting to parse '{conf_file}' as legacy ini file."
                )
                try:
                    logging.config.fileConfig(
                        str(conf_file),
                        disable_existing_loggers=False
                    )
                    logging.getLogger(__name__).info(
                        f"Successfully loaded '{conf_file}'"
                    )
                    return
                except Exception as ex:
                    if suffix == '.ini':
                        raise ex
                    else:
                        logging.getLogger(__name__).warning(
                            f"Unable to parse '{conf_file}' as legacy ini"
                        )

            if not config:
                raise ValueError(f"Unable to parse '{conf_file}'")

        logging.config.dictConfig(config)
        logging.getLogger(__name__).info(
            "Successfully loaded configuration"
        )



# def enable_std_out_logging(
#   log_name=None,
#   log_level=None,
#   log_formatter=None
# ):
#     log = logging.getLogger(log_name)

#     log_level = log_level if log_level else log.level
#     log_formatter = log_formatter if log_formatter else log_formatter

#     stdout_log = logging.StreamHandler()
#     stdout_log.setLevel(log_level)
#     stdout_log.setFormatter(log_formatter)
#     log.addHandler(stdout_log)


# def enable_file_logging(
#     log_file,
#     log_name=None,
#     log_level=None,
#     log_formatter=None
# ):
#     log = logging.getLogger(log_name)
#     log_level = log_level if log_level else log.level
#     log_formatter = log_formatter if log_formatter else log_formatter

#     file_log = logging.FileHandler(log_file)
#     file_log.setLevel(log_level)
#     file_log.setFormatter(log_formatter)
#     log.addHandler(file_log)



import time
import configargparse
import re
from pathlib import Path
import logging
import redgrease.formatting
import redgrease.loader

args = configargparse.ArgParser(
    description="Scans one or more directories for Redis Gears scripts, and "
    "executes them in a Redis Gears instance or cluster. "
    "Can optionally run continiously, montoring and re-loading scripts "
    "whenever changes are detected.",
    default_config_files=['./*.conf', '/etc/redgrease/conf.d/*.conf'],
)

args.add_argument(
    'directories',
    metavar='dir_path',
    nargs='+',
    type=Path,
    help="One or more directories containing Redis Gears scripts to watch",
    )

args.add_argument(
    '-c',
    '--config',
    env_var='CONFIG_FILE',
    metavar="PATH",
    is_config_file=True,
    help="Config file path",
)
args.add_argument(
    '--index-prefix',
    env_var='INDEX_PREFIX',
    metavar="PREFIX",
    required=False,
    type=str,
    default=redgrease.loader.default_index_prefix,
    help="Redis key prefix added to the index of monitored/executed script "
    "files.",
    )
args.add_argument(
    '-r',
    '--recursive',
    env_var='RECURSIVE',
    action='store_true',
    help="Recursively watch subdirectories.",
)
args.add_argument(
    '--script-pattern',
    env_var='SCRIPT_PATTERN',
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_script_pattern,
    help="File name pattern (glob-style) that must be matched for scripts to "
    "be loaded.",
)
args.add_argument(
    '--requirements-pattern',
    env_var="REQUIREMENTS_PATTERN",
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_requirements_pattern,
    help="File name pattern (glob-style) that must be matched for requirement "
    "files to be loaded.",
)
args.add_argument(
    "--unblocking-pattern",
    env_var="UNBLOCKING_PATTERN",
    metavar="PATTERN",
    required=False,
    type=re.compile,
    default=redgrease.loader.default_unblocking_pattern,
    help="Scripts with file paths that match this regular expression, "
    "will be executed with the 'UNBLOCKING' modifier, i.e. async execution."
)
args.add_argument(
    '-i',
    '--ignore',
    env_var="IGNORE",
    metavar='PATTERN',
    action='append',
    required=False,
    help="Ignore files matching this pattern.",
)
args.add_argument(
    '-w',
    '--watch',
    env_var='WATCH',
    metavar="SECONDS",
    nargs='?',
    type=float,
    default=0.0,
    const=5.0,
    help="If set, the directories will be continously montiored for updates/"
    "modifications to scripts and requirement files, and automatically loaded/"
    "rerun. The flag takes an optional value specifying the  duration, "
    "in seconds, to wait for further updates/modifications to files, before "
    "executing. "
    "This 'hysteresis' period is to prevent malformed scripts to be "
    "unnecessarily loaded during coding. "
    "If no value is supplied, the duration is defaulting to 5 seconds."
)
args.add_argument(
    '-s',
    '--server',
    env_var="SERVER",
    type=str,
    const='redis',
    default='localhost',
    nargs='?',
    help="Redis Gears host server IP or hostname."
)
args.add_argument(
    '-p',
    '--port',
    env_var='PORT',
    type=int,
    default=6379,
    help="Redis Gears host port number"
)
args.add_argument(
    '-l',
    '--log-config',
    env_var='LOG_CONFIG',
    type=str,
    help="",
)

config = args.parse_args()

redgrease.formatting.initialize_logger(config=config.log_config)

log = logging.getLogger(__name__)


def main():
    log.debug(config)
    loader = redgrease.loader.GearsLoader(
        script_pattern=config.script_pattern,
        requirements_pattern=config.requirements_pattern,
        unblocking_pattern=config.unblocking_pattern,
        ignore_patterns=config.ignore,
        index_prefix=config.index_prefix,
        server=config.server,
        port=config.port,
        observe=config.watch
    )
    for directory in config.directories:
        loader.add_directory(
            directory=directory,
            recursive=config.recursive
        )

    if loader.observer:
        log.info("Starting directory observer!")
        loader.start()
        running = True

        try:
            while running:
                time.sleep(1)
        except KeyboardInterrupt:
            running = False
            log.warning("Interrupted by user. Ending!")
        finally:
            loader.stop()


if __name__ == '__main__':
    main()



from threading import Timer
from typing import Callable, Hashable
import logging

log = logging.getLogger(__name__)


def nop(*args, **kwargs):
    """No Operation: Do nothing, regardless what is passed as arguments"""
    pass


class ResettableTimer:
    """Simple reresttable timer that will call a function when the timer elapses.
    Both the timer, as well as the function and its arguments can be reset
    when/if the timer is reset.
    """
    def __init__(self):
        """Create timer"""
        self.handler = nop
        self.handler_args = []
        self.handler_kwargs = {}
        self.timer = None

    def __del__(self):
        """Destructor. Cancels timer"""
        self.cancel()

    def cancel(self):
        """Cancel the timer (if set) so that it wont elapse"""
        if self.timer:
            try:
                self.timer.cancel()
            except Exception:
                pass
        self.timer = None

    def set(
        self,
        timeout: float,
        handler: Callable[..., None],
        *handler_args,
        **handler_kwargs
    ):
        """Reset the timer and callback

        Args:
            timeout (float): New timeout, in seconds.

            handler (Callable[..., None]): Handler function.

            Note that any additional positional and/or keyword arguments will
            be forwarded as arguments to the handler, when the timer elapses.
        """
        self.cancel()
        self.handler = handler
        self.handler_args = handler_args
        self.handler_kwargs = handler_kwargs
        self.timer = Timer(timeout if timeout else 1.0, self.trigger)
        self.timer.start()

    def trigger(self):
        """Trigger the handler immediately. Will cancel the timer."""
        self.cancel()
        if self.handler:
            self.handler(*self.handler_args, **self.handler_kwargs)


class HysteresisHandlerIndex:
    """
    Index of signal handlers, triggering each handler only when its'
    corresponding signal has not been registered for a set amount of time.
    This is useful for holding back some event handler as long as new events
    are still occuring, triggering only when the events stop.
    For example loading a file when there hasnt been any modifications to it
    for a period of time
    """
    def __init__(self, hysteresis_duration: float = 1.0):
        """Initialize the Index.

        Args:
            hysteresis_duration (float, optional): The required number of
            seconds without any signals for a given handler for it to trigger.
            Defaults to 1.0 second.
        """
        self.hysteresis_duration = hysteresis_duration
        self.handlers = {}

    def signal(
        self,
        handler_id: Hashable,
        handler_function: Callable[..., None],
        *handler_args,
        **handler_kwargs
    ):
        """Register a new signal for a specific handler.

        Args:
            handler_id ([Hashable]): Id of the handler / signal

            handler_function (Callable[..., None]): Signal handler function.

            Note that any additional positional and/or keyword arguments will
            be forwarded as arguments to the handler, when triggered.
        """
        if handler_id not in self.handlers:
            self.handlers[handler_id] = ResettableTimer()

        self.handlers[handler_id].set(
            self.hysteresis_duration,
            handler_function,
            *handler_args,
            **handler_kwargs
        )



import sys

if 'redisgears' in sys.modules:
    # Import the default functions and classes
    from redisgears import atomicCtx as atomic  # noqa: F401
    from redisgears import executeCommand as execute  # noqa: F401
    from redisgears import getMyHashTag as hashtag  # noqa: F401
    from redisgears import log  # noqa: F401
    from __main__ import gearsConfigGet  # noqa: F401
    from __main__ import GearsBuilder  # noqa: F401
    from __main__ import GB  # noqa: F401
else:
    # Import placeholder functions and classes
    from .placeholders import (   # noqa: F401
        atomic,
        execute,
        hashtag,
        log,
        configGet,
        gearsConfigGet,
        GearsBuilder,
        GB,
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import setuptools
import pathlib
import sys

here = pathlib.Path(__file__).parent.resolve()


def text(*names, encoding='utf8'):
    return here.joinpath(*names).read_text(encoding=encoding)


try:
    version = open('.version').readline()
except FileNotFoundError:
    print("Unable to determine version.")
    print("Set the version on the first line of a '.version' file")
    sys.exit(404)

client_extras = ['redis', 'attrs', ]
cli_extras = client_extras + ['watchdog', 'ConfigArgParse', 'pyyaml', ]

all_extras = cli_extras

setuptools.setup(
    name="redgrease",
    version=version,
    license="MIT License",
    description="RedisGears helper package",
    long_description=text("README.md"),
    long_description_content_type="text/markdown",

    author="Anders Åström",
    author_email="anders@lyngon.com",
    url="https://github.com/lyngon/redgrease",

    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    classifiers=[  # https://pypi.org/classifiers/
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Database",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Application Frameworks",  # noqa: E501
    ],
    install_requires=[],  # Should remain empty for minimal runtime install.
    extras_require={
        'all': all_extras,
        'client': client_extras,
        'cli':  cli_extras,
    },
    entry_points={
        'console_scripts': ['redgrease=redgrease.cli:main [cli]'],
    },
    setup_requires=[],
    python_requires='>=3.6',
    keywords="Redis, Gears, development"
)

sys.exit(0)
